﻿using System;
using System.Windows.Forms;
using System.Drawing;
using System.Threading;
using System.Reflection;
using System.IO;
using System.Runtime.InteropServices;
using RootPro.RootProCAD;
using RootPro.RootProCAD.Geometry;
using RootPro.RootProCAD.UI;

namespace RCAddInAdvancedZoomCS
{
	[System.AddIn.AddIn("AppAddIn", Version = "1.0", Publisher = "", Description = "")]
	public partial class AppAddIn
	{
		[DllImport("user32.dll", SetLastError = true, CharSet = CharSet.Auto)]
		private static extern int PostMessage(IntPtr hWnd, uint Msg, uint wParam, uint lParam);

		[DllImport("user32.dll", SetLastError = true, CharSet = CharSet.Auto)]
		private static extern void InvalidateRect(IntPtr hWnd, IntPtr lpRect, int bErase);

		private const uint WM_LBUTTONDOWN = 0x0201;
		private const uint WM_LBUTTONUP = 0x0202;
		private const uint WM_RBUTTONDOWN = 0x0204;
		private const uint WM_RBUTTONUP = 0x0205;
		private int MAKELPARAM(int x, int y) { return ((y << 16) | (x & 0xFFFF)); }

		private Point mouseDownPoint;
		private MouseButtons mouseButtons = MouseButtons.None;
		private bool startZoom = false;
		private bool panMode = false;

		private bool skipLDown = false;
		private bool skipRDown = false;

		private Point2d viewBottomLeftPoint;
		private Point2d viewTopRightPoint;

		private void AppAddIn_Startup(object sender, EventArgs e)
		{
			AppDomain.CurrentDomain.AssemblyResolve += new ResolveEventHandler(ResolveAssembly);

			DrawingViewMouseDown += new DrawingViewMouseEventHandler(application_DrawingViewMouseDown);
			DrawingViewMouseUp += new DrawingViewMouseEventHandler(application_DrawingViewMouseUp);
			DrawingViewMouseMove += new DrawingViewMouseEventHandler(application_DrawingViewMouseMove);
			DrawingViewMouseLeave += new DrawingViewMouseEventHandler(application_DrawingViewMouseLeave);
			DrawingViewDrawTemporary += new DrawingViewDrawTemporaryEventHandler(application_DrawingViewDrawTemporary);

			viewBottomLeftPoint = new Point2d();
			viewTopRightPoint = new Point2d();
		}

		private void AppAddIn_Shutdown(object sender, EventArgs e)
		{
			DrawingViewMouseDown -= application_DrawingViewMouseDown;
			DrawingViewMouseUp -= application_DrawingViewMouseUp;
			DrawingViewMouseMove -= application_DrawingViewMouseMove;
			DrawingViewMouseLeave -= application_DrawingViewMouseLeave;
			DrawingViewDrawTemporary -= application_DrawingViewDrawTemporary;
		}

		#region VSTA generated code
		private void InternalStartup()
		{
			this.Startup += new System.EventHandler(AppAddIn_Startup);
			this.Shutdown += new System.EventHandler(AppAddIn_Shutdown);
		}
		#endregion

		/// <summary>
		/// 作図画面上でマウスボタンを押したときに呼ばれる
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void application_DrawingViewMouseDown(object sender, DrawingViewMouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left || e.Button == MouseButtons.Right)
			{
				if (e.Button == MouseButtons.Left && skipLDown)
				{
					skipLDown = false;
					return;
				}
				if (e.Button == MouseButtons.Right && skipRDown)
				{
					skipRDown = false;
					return;
				}

				// 最初のボタンダウン
				if (mouseButtons == MouseButtons.None)
				{
					DrawingView drawingView = e.DrawingView;
					if ((drawingView.ModifierKeys & (Keys.Control | Keys.Shift | Keys.Alt)) == Keys.None)
					{
						mouseDownPoint = e.ClientPoint;
						mouseButtons = e.Button;
						e.Cancel = true;
					}
				}
				// 2個目のボタンダウン
				else if ((mouseButtons == MouseButtons.Left && e.Button == MouseButtons.Right) ||
					(mouseButtons == MouseButtons.Right && e.Button == MouseButtons.Left))
				{
					if (e.ClientPoint == mouseDownPoint)
					{
						mouseButtons = mouseButtons | e.Button;
						e.Cancel = true;

						// 左右両ボタンをクリックしたら画面コマンド開始
						if (mouseButtons == (MouseButtons.Left | MouseButtons.Right))
						{
							startZoom = true;

							DrawingView drawingView = e.DrawingView;

							// Ctrl キーを押している場合はリアルタイムパン開始
							if ((drawingView.ModifierKeys & Keys.Control) == Keys.Control)
							{
								viewBottomLeftPoint = drawingView.ViewBottomLeftPoint;
								viewTopRightPoint = drawingView.ViewTopRightPoint;

								drawingView.BeginPanning(e.ClientPoint);
								panMode = true;
							}

							// 再描画
							InvalidateRect(drawingView.WindowHandle, (IntPtr)0, 0);
						}
					}
					else
					{
						mouseButtons = MouseButtons.None;
					}
				}
			}
		}

		/// <summary>
		/// 作図画面上でマウスを動かしたときに呼ばれる
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void application_DrawingViewMouseMove(object sender, DrawingViewMouseEventArgs e)
		{
			DrawingView drawingView = e.DrawingView;

			// 左ボタン、右ボタンだけをドラッグした場合は、ボタンダウンメッセージを送信して終了
			if (!startZoom && (mouseButtons == MouseButtons.Left || mouseButtons == MouseButtons.Right))
			{
				if (e.ClientPoint != mouseDownPoint)
				{
					if (drawingView.MouseButtons == MouseButtons.Left || drawingView.MouseButtons == MouseButtons.Right)
					{
						if (drawingView.MouseButtons == MouseButtons.Left)
							skipLDown = true;
						else
							skipRDown = true;

						IntPtr hwnd = e.DrawingView.WindowHandle;
						PostMessage(hwnd, drawingView.MouseButtons == MouseButtons.Left ? WM_LBUTTONDOWN : WM_RBUTTONDOWN,
							(uint)0, (uint)MAKELPARAM(mouseDownPoint.X, mouseDownPoint.Y));
					}

					mouseButtons = MouseButtons.None;
					startZoom = false;
				}
			}
			// ボタンを押していない場合は終了
			else if (drawingView.MouseButtons == MouseButtons.None)
			{
				mouseButtons = MouseButtons.None;
				startZoom = false;
			}
			// リアルタイムパンの場合
			else if (startZoom && panMode)
			{
				drawingView.PanView(e.ClientPoint);
			}
		}

		/// <summary>
		/// 作図画面上でマウスボタンを離したときに呼ばれる
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void application_DrawingViewMouseUp(object sender, DrawingViewMouseEventArgs e)
		{
			// 押している左右ボタンを離した場合
			if (((e.Button == MouseButtons.Left) && ((mouseButtons & MouseButtons.Left) == MouseButtons.Left)) ||
				((e.Button == MouseButtons.Right) && ((mouseButtons & MouseButtons.Right) == MouseButtons.Right)))
			{
				mouseButtons = mouseButtons & ~e.Button;

				// 画面コマンドを開始していた場合
				if (startZoom)
				{
					// 左右両方のボタンを離した
					if (mouseButtons == MouseButtons.None)
					{
						DrawingView drawingView = e.DrawingView;

						// リアルタイムパン終了
						if (panMode)
						{
							drawingView.EndPanning();
							panMode = false;
						}
						// 右下ドラッグで 2 点拡大
						else if (e.ClientPoint.X > mouseDownPoint.X && e.ClientPoint.Y > mouseDownPoint.Y)
						{
							viewBottomLeftPoint = drawingView.ViewBottomLeftPoint;
							viewTopRightPoint = drawingView.ViewTopRightPoint;

							Point2d worldPoint1 = drawingView.TransformPointClientToWorld(e.ClientPoint);
							Point2d worldPoint2 = drawingView.TransformPointClientToWorld(mouseDownPoint);
							drawingView.ZoomWindow(worldPoint1, worldPoint2);
						}
						// 左上ドラッグで縮小
						else if (e.ClientPoint.X < mouseDownPoint.X && e.ClientPoint.Y < mouseDownPoint.Y)
						{
							viewBottomLeftPoint = drawingView.ViewBottomLeftPoint;
							viewTopRightPoint = drawingView.ViewTopRightPoint;

							drawingView.ZoomOut();
						}
						// 右上ドラッグで全図形表示
						else if (e.ClientPoint.X > mouseDownPoint.X && e.ClientPoint.Y < mouseDownPoint.Y)
						{
							viewBottomLeftPoint = drawingView.ViewBottomLeftPoint;
							viewTopRightPoint = drawingView.ViewTopRightPoint;

							drawingView.ZoomAll();
						}
						// 左下ドラッグで前回表示
						else if (e.ClientPoint.X < mouseDownPoint.X && e.ClientPoint.Y > mouseDownPoint.Y)
						{
							Point2d lastBottomLeft = viewBottomLeftPoint;
							Point2d lastTopRight = viewTopRightPoint;

							viewBottomLeftPoint = drawingView.ViewBottomLeftPoint;
							viewTopRightPoint = drawingView.ViewTopRightPoint;

							drawingView.ZoomWindow(lastBottomLeft, lastTopRight);
						}
						// 同じ点のクリックで中心表示
						else if (e.ClientPoint.X == mouseDownPoint.X && e.ClientPoint.Y == mouseDownPoint.Y)
						{
							viewBottomLeftPoint = drawingView.ViewBottomLeftPoint;
							viewTopRightPoint = drawingView.ViewTopRightPoint;

							Point2d worldPoint = drawingView.TransformPointClientToWorld(e.ClientPoint);
							drawingView.ScrollTo(worldPoint);
						}

						startZoom = false;
					}

					e.Cancel = true;
				}
				else
				{
					// 同じ位置で左クリック or 右クリックした場合
					if (mouseButtons == MouseButtons.None && e.ClientPoint == mouseDownPoint)
					{
						// ボタンダウン、ボタンアップメッセージを送信する
						IntPtr hwnd = e.DrawingView.WindowHandle;
						PostMessage(hwnd, (e.Button == MouseButtons.Left) ? WM_LBUTTONDOWN : WM_RBUTTONDOWN,
							(uint)0, (uint)MAKELPARAM(e.ClientPoint.X, e.ClientPoint.Y));
						PostMessage(hwnd, (e.Button == MouseButtons.Left) ? WM_LBUTTONUP : WM_RBUTTONUP,
							(uint)0, (uint)MAKELPARAM(e.ClientPoint.X, e.ClientPoint.Y));

						if (e.Button == MouseButtons.Left)
							skipLDown = true;
						else
							skipRDown = true;

						e.Cancel = true;
					}
				}
			}
		}

		/// <summary>
		/// 作図画面外にマウスが移動したときに呼ばれる
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void application_DrawingViewMouseLeave(object sender, DrawingViewMouseEventArgs e)
		{
			mouseButtons = MouseButtons.None;
			startZoom = false;
		}

		/// <summary>
		/// 作図画面上に一時的な描画を実行するときに呼ばれる
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void application_DrawingViewDrawTemporary(object sender, DrawingViewDrawTemporaryEventArgs e)
		{
			DrawingView drawingView = e.DrawingView;
			if (startZoom)
			{
				if(!panMode)
				{
					int x = Math.Min(e.ClientPoint.X, mouseDownPoint.X);
					int y = Math.Min(e.ClientPoint.Y, mouseDownPoint.Y);
					System.Drawing.Point point = new System.Drawing.Point(x, y);
					int width = Math.Abs(e.ClientPoint.X - mouseDownPoint.X);
					int height = Math.Abs(e.ClientPoint.Y - mouseDownPoint.Y);

					if(width > 0 || height > 0)
						e.DrawRectangle(point, width, height, Color.Red, 1, System.Drawing.Drawing2D.DashStyle.Solid, null);

					// マウスカーソルを右下に移動
					if (e.ClientPoint.X > mouseDownPoint.X && e.ClientPoint.Y > mouseDownPoint.Y)
					{
						e.DrawToolTipText(e.ClientPoint, Properties.Resources.FunctionZoomRange, 20);
					}
					// マウスカーソルを左上に移動
					else if (e.ClientPoint.X < mouseDownPoint.X && e.ClientPoint.Y < mouseDownPoint.Y)
					{
						e.DrawToolTipText(e.ClientPoint, Properties.Resources.FunctionZoomOut, 20);
					}
					// マウスカーソルを右上に移動
					else if (e.ClientPoint.X > mouseDownPoint.X && e.ClientPoint.Y < mouseDownPoint.Y)
					{
						e.DrawToolTipText(e.ClientPoint, Properties.Resources.FunctionZoomAll, 20);
					}
					// マウスカーソルを左下に移動
					else if (e.ClientPoint.X < mouseDownPoint.X && e.ClientPoint.Y > mouseDownPoint.Y)
					{
						e.DrawToolTipText(e.ClientPoint, Properties.Resources.FunctionLastView, 20);
					}
					// マウスカーソルは同じ座標
					else if (e.ClientPoint.X == mouseDownPoint.X && e.ClientPoint.Y == mouseDownPoint.Y)
					{
						e.DrawToolTipText(e.ClientPoint, Properties.Resources.FunctionScrollTo, 20);
					}
				}
			}
		}

		private Assembly ResolveAssembly(object sender, ResolveEventArgs args)
		{
			AssemblyName assemblyName = new AssemblyName(args.Name);

			foreach (Assembly assembly in AppDomain.CurrentDomain.GetAssemblies())
			{
				if (AssemblyName.ReferenceMatchesDefinition(assemblyName, assembly.GetName()))
					return assembly;
			}

			if (assemblyName.Name.EndsWith(".resources"))
			{
				if (assemblyName.CultureName.Equals(Thread.CurrentThread.CurrentUICulture.Name, StringComparison.OrdinalIgnoreCase))
				{
					string dllPath = Path.GetDirectoryName(this.AddInFullPath);
					dllPath = Path.Combine(dllPath, assemblyName.CultureName);
					dllPath = Path.Combine(dllPath, assemblyName.Name + ".dll");

					if (File.Exists(dllPath))
					{
						try
						{
							return Assembly.LoadFrom(dllPath);
						}
						catch
						{
						}
					}
				}
			}

			return null;
		}

	}
}
